''' Test import from import data saved with XIO-sensors, through subclassing "IMU_Base" '''

# Author: Thomas Haslwanter

import os
import numpy as np

import unittest
from time import sleep
from skinematics.sensors.xio import XIO

myPath = os.path.dirname(os.path.abspath(__file__))

class TestSequenceFunctions(unittest.TestCase):

    def test_import_xio(self):
        # Get data, with a specified input from an XIO system
        in_file = os.path.join(myPath, 'data', 'data_xio')
        sensor = XIO(in_file=in_file, q_type=None)

        rate = sensor.rate
        acc = sensor.acc
        omega = sensor.omega

        self.assertAlmostEqual((rate - 256), 0)
        self.assertAlmostEqual( (np.rad2deg(omega[0,2]) -10.125), 0)

if __name__ == '__main__':
    unittest.main()
    print('Thanks for using programs from Thomas!')
    sleep(0.2)



'''
Import data saved with "x-IMU3" sensors from X-IO, through subclassing "IMU_Base"
Note that the data are in three files: data come from

    - an inertial sensor
    - a magnetometer
    - a high-speed accelerometer

More info about the sensor on https://x-io.co.uk/x-imu3/
'''

# Author: Thomas Haslwanter

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import constants
import os

from skinematics.imus import IMU_Base


def read_datafiles(in_files, rate):
    '''Read data from XIO3-files.
    To provide a fixed frequency, the data are re-sampled to "rate".

    Parameters
    ----------

    in_file : list with file-paths of the data-files
    rate : sample-rate for the interpolated data

    Returns
    -------
    out_list: list
            Contains the following parameters:

            - time [s]
            - acceleration [g]
            - angular_velocity [deg/s]
            - mag_field_direction [uT]
    '''

    data_imu = pd.read_csv(in_files['IMU']).values
    data_b = pd.read_csv(in_files['mag']).values
    # df_acc = pd.read_csv(in_files['HGacc'])

    t_i = data_imu[:, 0]
    t_b = data_b[:, 0]
    # t_acc = df_acc['Timestamp (us)']

    start = np.max([t_i[0], t_b[0]])
    stop = np.min([t_i[-1], t_b[-1]])
    step = 1e6 / rate   # since the timestamps are in micro-seconds

    # Interpolate the IMU-data
    t = np.arange(start, stop, step)
    data = []
    for col in range(1, 7):
        data.append(np.interp(t, t_i, data_imu[:, col]))

    # Interpolate the magnetic field measurements
    for col in range(1, 4):
        data.append(np.interp(t, t_b, data_b[:, col]))

    data = np.array(data).T
    acc = data[:, 3:6]
    omega = data[:, :3]
    b = data[:, 6:]

    return (acc, omega, b)


class XIO3(IMU_Base):
    """Concrete class based on abstract base class IMU_Base """

    def get_data(self, in_file, in_data):

        '''
        Get the recorded data, and assign them to the corresponding attributes
        of "self". For the XIO3, the data are stored in different files:

        - "Inertial.csv" : data from the inertial sensor
        - "Magnetometer.csv" : data from the magnetometer
        - "HighGAccelerometer.csv" : data from the high-speed-accelerometer

        For the protocoll here, data from the HighGAccelerometer are ignored,
        data from the magnetometer are interpolated to provide the same frequency
        as the data from the inertial sensor.

        Assigns the following properties:
            - time : recording, starting at 0 (sec)
            - acc : acceleration (g)
            - omega : angular_velocity (deg/s)
            - mag : mag_field_direction ()

        Parameters
        ----------
        in_selection : string
                Directory containing all the data-files, or
                filename of one file in that directory
        in_data : requested sample rate for interpolation


        '''

        in_selection = in_file
        rate = in_data

        if os.path.isdir(in_selection):
            in_dir = in_selection
        else:
            in_file = in_selection
            in_dir = os.path.split(in_file)[0]

        file_list = os.listdir(in_dir)

        # Get the filenames, based on the XIO3-definitions
        files = {}
        for file in file_list:
            if file.find('Inertial') >= 0:
                files['IMU'] = os.path.join(in_dir, file)
            if file.find('Magnetometer') >= 0:
                files['mag'] = os.path.join(in_dir, file)
            if file.find('HighGAccelerometer') >= 0:
                files['HGacc'] = os.path.join(in_dir, file)

        # Read the sensor-data
        acc, omega, b = read_datafiles(files, rate)

        # Set the class properties
        in_data = {'rate': rate,
               'acc':   acc * constants.g,
               'omega': np.deg2rad(omega),
               'mag':   b}
        self._set_data(in_data)


if __name__ == '__main__':
    test_dir = r'..\..\tests\data\data_xio3'
    assert os.path.exists(test_dir)

    sample_rate = 50. # Hz
    my_sensor = XIO3(in_file=test_dir, in_data=sample_rate)

    plt.plot(my_sensor.acc)
    print(my_sensor.rate)
    plt.show()
    print('Done')




# from .context import skinematics

import unittest
import numpy as np
from numpy import sin, cos, array, r_, vstack, abs, tile, pi
from numpy.linalg import norm
import os
from skinematics import imus, quat, vector, rotmat
from time import sleep
from skinematics.simulations.simulate_movements import simulate_imu

myPath = os.path.dirname(os.path.abspath(__file__))

class TestSequenceFunctions(unittest.TestCase):
    def setUp(self):

        # Those are currently not needed
        # self.qz  = r_[cos(0.1), 0,   0,    sin(0.1)]
        # self.qy  = r_[cos(0.1), 0, sin(0.1), 0]
        # self.quatMat = vstack((self.qz,self.qy))
        # self.q3x = r_[sin(0.1),  0,     0]
        # self.q3y = r_[  0,   sin(0.1), 0]
        # self.delta = 1e-4

        # Simulate IMU-data
        duration_movement = 1  # [sec]
        duration_total = 1  # [sec]
        rate = 100  # [Hz]

        B0 = vector.normalize([1, 0, 1])  # geomagnetic field, re earth

        rotation_axis = [0, 1, 0]
        angle = 90

        translation = [1, 0, 0]
        distance = 0

        self.q_init = [0, 0, 0]
        self.pos_init = [0, 0, 0]

        self.imu_signals, self.body_pos_orient = simulate_imu(
            rate,
            duration_movement,
            duration_total,
            q_init=self.q_init,
            rotation_axis=rotation_axis,
            deg=angle,
            pos_init=self.pos_init,
            direction=translation,
            distance=distance,
            B0=B0,
        )

    def test_analytical(self):

        # Analyze the simulated data with "analytical"
        imu = self.imu_signals

        q, pos, vel = imus.analytical(
            R_initialOrientation=np.eye(3),
            omega=imu["omega"],
            initialPosition=np.zeros(3),
            accMeasured=imu["gia"],
            rate=imu["rate"],
        )

        # and then check, if the position is [0,0,0], and the orientation-quat = [0, sin(45), 0]
        self.assertTrue(np.max(np.abs(pos[-1])) < 0.001)  # less than 1mm

        result = quat.q_vector(q[-1])
        correct = array([0.0, np.sin(np.deg2rad(45)), 0.0])
        error = norm(result - correct)
        self.assertAlmostEqual(error, 0)

    # def test_kalman(self):

    #     # Analyze the simulated data with "kalman"
    #     imu = self.imu_signals
    #     q_kalman = imus.kalman(imu["rate"], imu["gia"], imu["omega"], imu["magnetic"])

    #     # and then check, if the quat_vector = [0, sin(45), 0]
    #     result = quat.q_vector(q_kalman[-1])  # [0, 0.46, 0]
    #     correct = array([0.0, np.sin(np.deg2rad(45)), 0.0])  # [0, 0.71, 0]
    #     error = norm(result - correct)
    #     self.assertAlmostEqual(
    #         error, 0, places=2
    #     )  # It is not clear why the Kalman filter is not more accurate

    #     # Get data
    #     inFile = os.path.join(myPath, "data", "data_xsens.txt")
    #     from skinematics.sensors.xsens import XSens

    #     initialPosition = array([0, 0, 0])
    #     R_initialOrientation = rotmat.R(0, 90)

    #     sensor = XSens(
    #         in_file=inFile,
    #         R_init=R_initialOrientation,
    #         pos_init=initialPosition,
    #         q_type="kalman",
    #     )
    #     print(sensor.source)
    #     q = sensor.quat

    def test_madgwick(self):

        from skinematics.sensors.manual import MyOwnSensor

        ## Get data
        imu = self.imu_signals
        in_data = {
            "rate": imu["rate"],
            "acc": imu["gia"],
            "omega": imu["omega"],
            "mag": imu["magnetic"],
        }

        my_sensor = MyOwnSensor(
            in_file="Simulated sensor-data",
            in_data=in_data,
            R_init=quat.convert(self.q_init, to="rotmat"),
            pos_init=self.pos_init,
            q_type="madgwick",
        )

        # and then check, if the quat_vector = [0, sin(45), 0]
        q_madgwick = my_sensor.quat

        result = quat.q_vector(q_madgwick[-1])
        correct = array([0.0, np.sin(np.deg2rad(45)), 0.0])
        error = norm(result - correct)

        # self.assertAlmostEqual(error, 0)
        self.assertTrue(error < 1e-3)

        ##inFile = os.path.join(myPath, 'data', 'data_xsens.txt')
        ##from skinematics.sensors.xsens import XSens

        ##initialPosition = array([0,0,0])
        ##R_initialOrientation = rotmat.R(0,90)

        ##sensor = XSens(in_file=inFile, R_init = R_initialOrientation, pos_init = initialPosition, q_type='madgwick')
        ##q = sensor.quat

    def test_mahony(self):

        from skinematics.sensors.manual import MyOwnSensor

        ## Get data
        imu = self.imu_signals
        in_data = {
            "rate": imu["rate"],
            "acc": imu["gia"],
            "omega": imu["omega"],
            "mag": imu["magnetic"],
        }

        my_sensor = MyOwnSensor(
            in_file="Simulated sensor-data",
            in_data=in_data,
            R_init=quat.convert(self.q_init, to="rotmat"),
            pos_init=self.pos_init,
            q_type="mahony",
        )

        # and then check, if the quat_vector = [0, sin(45), 0]
        q_mahony = my_sensor.quat

        result = quat.q_vector(q_mahony[-1])
        correct = array([0.0, np.sin(np.deg2rad(45)), 0.0])
        error = norm(result - correct)

        # self.assertAlmostEqual(error, 0)
        self.assertTrue(error < 1e-3)

        ### Get data
        ##inFile = os.path.join(myPath, 'data', 'data_xsens.txt')
        ##from skinematics.sensors.xsens import XSens

        ##initialPosition = array([0,0,0])
        ##R_initialOrientation = rotmat.R(0,90)

        ##sensor = XSens(in_file=inFile, R_init = R_initialOrientation, pos_init = initialPosition, q_type='mahony')
        ##q = sensor.quat

    def test_IMU_calc_orientation_position(self):
        """Currently, this only tests if the two functions are running through"""

        # Get data, with a specified input from an XSens system
        # data_dir = resources.files('data')
        # inFile = data_dir/'data_xsens.txt'

        inFile = os.path.join(myPath, "data", "data_xsens.txt")
        initial_orientation = np.array([[1, 0, 0], [0, 0, -1], [0, 1, 0]])
        initial_position = np.r_[0, 0, 0]

        from skinematics.sensors.xsens import XSens

        sensor = XSens(
            in_file=inFile, R_init=initial_orientation, pos_init=initial_position
        )
        sensor.calc_position()
        print("done")

    def test_set_qtype(self):
        """Tests if the test crashes on any of the existing qtype options"""

        # Get data
        # data_dir = resources.files('data')
        # inFile = data_dir/'data_xsens.txt'
        inFile = os.path.join(myPath, "data", "data_xsens.txt")
        from skinematics.sensors.xsens import XSens

        initialPosition = array([0, 0, 0])
        R_initialOrientation = rotmat.R(0, 90)

        sensor = XSens(
            in_file=inFile,
            R_init=R_initialOrientation,
            pos_init=initialPosition,
            q_type="kalman",
        )

        allowed_values = ["analytical", "kalman", "madgwick", "mahony", None]

        for sensor_type in allowed_values:
            print("{0} is running".format(sensor_type))
            sensor.set_qtype(sensor_type)


if __name__ == "__main__":
    suite = unittest.TestSuite()
    suite.addTest(TestSequenceFunctions(methodName="test_set_qtype"))
    runner = unittest.TextTestRunner()
    runner.run(suite)

    # unittest.main()

    print("Thanks for using programs from Thomas!")
    sleep(0.2)



'''
Import data saved with "x-IMU" sensors from x-io, through subclassing "IMU_Base"
Note that the data are in two files:

    - a data-file
    - a reg-file

More info about the sensor on http://x-io.co.uk
'''

# Author: Thomas Haslwanter

import numpy as np
import pandas as pd
import os
from scipy import constants
import abc

from skinematics.imus import IMU_Base

def read_ratefile(reg_file):
    '''Read send-rates from an XIO sensor.
    "Disabled" channels have the "rate" set to "None".

    Parameters
    ----------
    in_file : string
            Has to be the "Registers"-file.

    Returns
    -------
    rates: directory
            Contains the send-rates for the different "params".
    '''

    params = ['Sensor',
              'DateTime',
              'BatteryAndThermometer',
              'InertialAndMagnetic',
              'Quaternion'
            ]

    rates = {}

    # The the file content
    with open(reg_file, 'r') as in_file:
        lines = in_file.readlines()

    # Get the send rates
    for param in params:
        for line in lines:
            if line.find(param) > 0:
                rate_flag = int(line.split(',')[2])
                if rate_flag:
                    '''
                    0 ... 1 Hz
                    1 ... 2 Hz
                    10 ... 512 Hz
                    '''
                    rates[param] = 2 ** (rate_flag-1)
                else:
                    # Disabled
                    rates[param] = None

    return rates

def read_datafile(in_file):
    '''Read data from an XIO "CalInertialAndMag"-file.

    Parameters
    ----------
    in_file : string
            Has to be the name of the "CalInertialAndMag"-file.

    Returns
    -------
    out_list: list
            Contains the following parameters:

            - acceleration
            - angular_velocity
            - mag_field_direction
            - packet_nr
    '''

    data = pd.read_csv(in_file)
    out_list = []

    # Extract the columns that you want, by name
    param_list=['Acc', 'Gyr', 'Mag', 'Packet']
    for Expression in param_list:
        out_list.append(data.filter(regex=Expression).values)

    return out_list

class XIO(IMU_Base):
    """Concrete class based on abstract base class IMU_Base """

    def get_data(self, in_file, in_data=None):
        '''
        Get the sampling rate, as well as the recorded data,
        and assign them to the corresponding attributes of "self".

        Assigns the following properties
            - rate : rate
            - acc : acceleration
            - omega : angular_velocity
            - mag : mag_field_direction
            - packet_nr : packet_nr

        Parameters
        ----------
        in_selection : string
                Directory containing all the data-files, or
                filename of one file in that directory
        in_data : not used here

        '''
        in_selection = in_file
        if os.path.isdir(in_selection):
            in_dir = in_selection
        else:
            in_file = in_selection
            in_dir = os.path.split(in_file)[0]

        file_list = os.listdir(in_dir)

        # Get the filenames, based on the XIO-definitions
        files = {}
        for file in file_list:
            if file.find('Registers') > 0:
                files['register'] = os.path.join(in_dir, file)
            if file.find('CalInertialAndMag') > 0:
                files['data'] = os.path.join(in_dir, file)

        # Read in the registers-file, and extract the sampling rates
        rates = read_ratefile(files['register'])

        # Read the sensor-data
        data  = read_datafile(files['data'])

        # Set the class properties
        in_data = {'rate':rates['InertialAndMagnetic'],
               'acc':   data[0] * constants.g,
               'omega': np.deg2rad(data[1]),
               'mag':   data[2]}
        self._set_data(in_data)

        self.packet_nr = data[-1]

if __name__ == '__main__':
    test_dir = r'../../tests/data/data_xio'
    assert os.path.exists(test_dir)

    my_sensor = XIO(in_file=test_dir)

    import matplotlib.pyplot as plt

    plt.plot(my_sensor.acc)
    print(my_sensor.rate)
    plt.show()
    print('Done')




''' Test manual data entry, through subclassing "IMU_Base" '''

# Author: Thomas Haslwanter

import unittest
from time import sleep
import os
from skinematics.sensors.xsens import XSens
from skinematics.sensors.manual import MyOwnSensor

myPath = os.path.dirname(os.path.abspath(__file__))

class TestSequenceFunctions(unittest.TestCase):

    def test_import_manual(self):
        # Get data, with a specified input from an XSens system
        in_file = os.path.join(myPath, 'data', 'data_xsens.txt')

        sensor = XSens(in_file=in_file, q_type=None)

        transfer_data = {'rate':sensor.rate,
                   'acc': sensor.acc,
                   'omega':sensor.omega,
                   'mag':sensor.mag}
        my_sensor = MyOwnSensor(in_file='My own 123 sensor.', in_data=transfer_data)

        self.assertEqual(my_sensor.rate, 50.)
        self.assertAlmostEqual( (my_sensor.omega[0,2] - 0.050860000000000002), 0)


    #def test_pos(self):
        ## suggested in github


if __name__ == '__main__':
    unittest.main()
    print('Thanks for using programs from Thomas!')
    sleep(0.1)



""" Test import from import data saved with XIO3 sensors from x-io,
through subclassing 'IMU_Base' """

# Author: Thomas Haslwanter

import numpy as np
import os

import unittest
from time import sleep
from skinematics.sensors.xio3 import XIO3

myPath = os.path.dirname(os.path.abspath(__file__))

class TestSequenceFunctions(unittest.TestCase):

    def test_import_xio(self):
        # Get data, with a specified input from an XIO system
        in_file = os.path.join(myPath, 'data', 'data_xio3')
        sensor = XIO3(in_file=in_file, q_type=None, in_data=50.)

        rate = sensor.rate
        acc = sensor.acc
        omega = sensor.omega

        self.assertAlmostEqual((rate - 50), 0)
        self.assertAlmostEqual(
            (np.rad2deg(omega[0, 2]) + 0.014149103923330339), 0)

if __name__ == '__main__':
    unittest.main()
    print('Thanks for using programs from Thomas!')
    sleep(0.2)



import unittest
from importlib import import_module

class TestSequenceFunctions(unittest.TestCase):
    
    def test_skinematics(self):
        for module in ['imus', 'markers', 'quat', 'rotmat', 'vector', 'view']:
            print(dir(module))
            full_name = 'skinematics.' + module
            import_module(full_name)



'''
Import data saved with "NGIMU" sensors from x-io, through subclassing "IMU_Base"
Note that the data are in two files:

    - a data-file
    - a reg-file

More info about the sensor on http://x-io.co.uk
'''

# Author: Thomas Haslwanter

import numpy as np
import pandas as pd
import os
from scipy import constants
import abc

from skinematics.imus import IMU_Base

def read_ratefile(reg_file):
    '''Read send-rates from an NGIMU sensor.
    "Disabled" channels have the "rate" set to "None".

    Parameters
    ----------
    in_file : string
            Has to be the "Registers"-file.

    Returns
    -------
    rates: directory
            Contains the send-rates for the following paramters:
            - sensors
            - magnitudes
            - quaternion
            - matrix
            - euler
            - linear
            - earth
            - altitude
            - temperature
            - humidity
            - battery
            - analogue
            - rssi

    '''

    rates = {}

    # The the file content
    with open(reg_file, 'r') as in_file:
        lines = in_file.readlines()

    # Get the send rates
    for line in lines:
        if line.find('rate') > 0:
            # e.g.: "/rate/quaternion, 50f"
            param_full, val_txt = line.split()
            value = np.float64(val_txt[:-1])  # skip the "f"
            param = param_full.split('/')[2][:-1]
            if value > 0:
                rates[param] = value
            else:
                rates[param] = None

    return rates

def read_datafile(in_file):
    '''Read data from an XIO "CalInertialAndMag"-file.

    Parameters
    ----------
    in_file : string
            Has to be the name of the "sensors.csv"-file.

    Returns
    -------
    out_list: list
            Contains the following parameters:

            - time [s]
            - acceleration [g]
            - angular_velocity [deg/s]
            - mag_field_direction [uT]
            - barometer [hPa]
    '''

    data = pd.read_csv(in_file)
    out_list = []

    # Extract the columns that you want, by name
    param_list=['Time', 'Acc', 'Gyr', 'Mag', 'Baro']
    for Expression in param_list:
        out_list.append(data.filter(regex=Expression).values)

    return out_list

class NGIMU(IMU_Base):
    """Concrete class based on abstract base class IMU_Base """

    def get_data(self, in_file, in_data=None):
        '''
        Get the sampling rate, as well as the recorded data,
        and assign them to the corresponding attributes of "self".
        For the x-io.NGIMU, the data are stored in different files:

        - "Settings.txt" : contains all the sensor-settings
        - "sensors.csv" : contains the recorded sensor data

        Assigns the following properties:
            - rate : rate
            - acc : acceleration
            - omega : angular_velocity
            - mag : mag_field_direction

        Parameters
        ----------
        in_selection : string
                Directory containing all the data-files, or
                filename of one file in that directory
        in_data : not used here


        '''

        in_selection = in_file
        if os.path.isdir(in_selection):
            in_dir = in_selection
        else:
            in_file = in_selection
            in_dir = os.path.split(in_file)[0]

        file_list = os.listdir(in_dir)

        # Get the filenames, based on the XIO-definitions
        files = {}
        for file in file_list:
            if file.find('Settings') >= 0:
                files['settings'] = os.path.join(in_dir, file)
            if file.find('sensors') >= 0:
                files['data'] = os.path.join(in_dir, file)

        # Read in the registers-file, and extract the sampling rates
        rates = read_ratefile(files['settings'])

        # Read the sensor-data
        data  = read_datafile(files['data'])

        # Set the class properties
        in_data = {'rate':rates['sensors'],
               'acc':   data[1] * constants.g,
               'omega': np.deg2rad(data[2]),
               'mag':   data[3]}
        self._set_data(in_data)


if __name__ == '__main__':
    test_dir = r'../../tests/data/data_ngimu'
    assert os.path.exists(test_dir)

    my_sensor = NGIMU(in_file=test_dir)

    import matplotlib.pyplot as plt

    plt.plot(my_sensor.acc)
    print(my_sensor.rate)
    plt.show()
    print('Done')




'''
"scikit-kinematics" primarily contains functions for working with 3D kinematics. (i.e.
quaternions and rotation matrices).

Compatible Python 3.

Dependencies
------------
numpy, scipy, matplotlib, pandas, sympy, pygame, pyOpenGL

Homepage
--------
http://work.thaslwanter.at/skinematics/html/

Copyright (c) 2024 Thomas Haslwanter <thomas.haslwanter@fh-ooe.at>

'''

import importlib

__author__ = "Thomas Haslwanter <thomas.haslwanter@fh-linz.at"
__license__ = "BSD 2-Clause License"
__version__ = "0.10.0"


required_imports = ['markers', 'quat', 'rotmat', 'vector', 'sensors']
optional_imports = ['imus', 'misc']

__all__ = []
for _m in required_imports:
    importlib.import_module('.'+_m, package='skinematics')
    __all__.append(_m)

for _m in optional_imports:
    try:
        importlib.import_module('.'+_m, package='skinematics')
        __all__.append(_m)
    except ModuleNotFoundError:
        print(f"Failed to import optional module {_m}. Install optional dependencies")

